from dataclasses import dataclass, field
from .enums.sum_types import SumTypes
from .enums.time_periods import TimePeriods
from ...services.locale.querytypes.queryservice import QueryService


@dataclass(frozen=False)
class QueryProperties:
    sumtype: SumTypes
    timecalc: TimePeriods
    cycle: TimePeriods
    queryservice: QueryService = field(default_factory=lambda: QueryService())



from dataclasses import dataclass, field
from .savings_status import SavingsStatus
from datetime import date, datetime
from .consumption_model import ConsumptionModel
import logging

_LOGGER = logging.getLogger(__name__)


@dataclass
class SavingsModel(ConsumptionModel):
    peak_price_per_kwh: float = 0
    car_connected_at: datetime | None = None
    prices: dict[date, list[float]] = field(default_factory=lambda: {})
    status: SavingsStatus = field(default_factory=lambda: SavingsStatus.Off)

    async def async_reset(self) -> None:
        self.status = SavingsStatus.Off
        self.prices = {}
        self.consumption = {}
        # self.peaks = {}

    async def async_re_initialize(self, incoming: dict):
        self.car_connected_at = incoming.get("car_connected_at", None)
        self.prices = incoming.get("prices", {})
        self.consumption = incoming.get("consumption", {})
        if all(
            [
                self.car_connected_at is not None,
                len(self.prices) > 0,
                len(self.consumption) > 0,
            ]
        ):
            self.status = SavingsStatus.Collecting
        else:
            _LOGGER.warning("Cannot re-initialize SavingsModel, missing data")

    async def async_add_prices(
        self, prices: list[float], _date: date | None = None
    ) -> None:
        if _date is None:
            _date = datetime.now().date()
        self.prices[_date] = prices

    async def async_add_to_consumption(
        self,
        consumption: float,
        date: date | None = None,
        hour: int | None = None,
    ) -> None:
        if self.status is not SavingsStatus.Collecting:
            return
        await self.async_add_consumption(consumption, date, hour)



from .models.stop_string import AllowanceObj, set_allowance_obj
from .models.datetime_model import DateTimeModel
from .models.hour_price import HourPrice
from .models.hourselection_model import HourSelectionModel
from statistics import stdev, mean, variance
from datetime import date, datetime, timedelta
from ...models.hourselection.hourselection_options import HourSelectionOptions
from .const import TODAY, TOMORROW
from .hourselection_calculations import (
    normalize_prices,
    deviation_from_mean,
)
from .permittance import set_initial_permittance, set_scooped_permittance
from .max_min_charge import MaxMinCharge


class HourSelectionService:
    def __init__(self, options: HourSelectionOptions = HourSelectionOptions()):
        self.options = options
        self.dtmodel = DateTimeModel()
        self.model = HourSelectionModel()
        self.max_min = MaxMinCharge(service=self, min_price=self.options.min_price)
        self._offset_dict: dict[datetime, dict] = {}

    def update(self):
        for hp in self.model.hours_prices:
            hp.set_passed(self.dtmodel)

    async def async_update(self):
        self.update()

    @property
    def all_hours(self) -> list[HourPrice]:
        self.update()
        return self.model.hours_prices

    @property
    def future_hours(self) -> list[HourPrice]:
        self.update()
        ret = [hp for hp in self.model.hours_prices if not hp.passed]
        if self.max_min.active:
            for r in ret:
                r.permittance = 0.0 if r.dt in self.max_min.non_hours else 1.0
        return ret

    @property
    def passed_hours(self) -> list[HourPrice]:
        self.update()
        return [hp for hp in self.model.hours_prices if hp.passed]

    @property
    def stopped_string(self) -> str:
        return self.allowance.display_name

    @property
    def allowance(self) -> AllowanceObj:
        return set_allowance_obj(self.dtmodel, self.future_hours)

    @property
    def average_kwh_price(self) -> float:
        return round(mean([hp.permittance * hp.price for hp in self.future_hours]), 2)

    @property
    def non_hours(self) -> list[datetime]:
        return [
            hp.dt
            for hp in self.model.hours_prices
            if not hp.passed and hp.permittance == 0.0
        ]

    @property
    def caution_hours(self) -> list[datetime]:
        return list(self.dynamic_caution_hours.keys())

    @property
    def dynamic_caution_hours(self) -> dict[datetime, float]:
        return {
            hp.dt: hp.permittance
            for hp in self.model.hours_prices
            if not hp.passed and 0.0 < hp.permittance < 1.0
        }

    @property
    def offset_dict(self) -> dict:
        return self._get_offset_dict()

    async def async_update_prices(
        self, prices: list[float], prices_tomorrow: list[float] = []
    ):
        self.model.prices_today = prices  # clean first
        self.model.prices_tomorrow = prices_tomorrow  # clean first
        if self._do_recalculate_prices(prices):
            self.model.hours_prices = await self.async_create_prices(
                prices, prices_tomorrow
            )

    async def async_update_adjusted_average(self, adjusted_average: float) -> None:
        self.model.adjusted_average = adjusted_average
        self.update()
        if len(self.model.hours_prices) > 0:
            await self.async_set_permittance(self.model.hours_prices)

    def _do_recalculate_prices(self, prices) -> bool:
        if [
            hp.price for hp in self.model.hours_prices if hp.day == self.dtmodel.hdate
        ] == prices:
            return False
        return True

    async def async_create_prices(
        self, prices: list[float], prices_tomorrow: list[float] = []
    ) -> list:
        # todo: fix to allow 23, 24,25, 92, 96, 100 for dst-dates.
        match len(prices):
            case 23 | 24 | 25:
                return await self.async_create_hour_prices_hourly(
                    prices, prices_tomorrow
                )
            case 92 | 96 | 100:
                return await self.async_create_hour_prices_quarterly(
                    prices, prices_tomorrow
                )
            case 0:
                return []
        raise ValueError(
            f"Length of pricelist must be either 23,24,25,92,96 or 100. Your length is {len(prices)}"
        )

    def _is_passed(self, hour, quarter) -> bool:
        if self.dtmodel.hour > hour:
            return True
        elif self.dtmodel.hour == hour:
            if self.dtmodel.quarter >= quarter:
                return True
        return False

    async def async_create_hour_prices_quarterly(
        self, prices: list[float], prices_tomorrow: list[float] = []
    ) -> list:
        # todo: handle here first if prices or prices_tomorrow are 92 or 100 in length (dst shift)
        ret = []
        for idx, p in enumerate(prices):
            assert isinstance(p, (float, int))
            hour = int(idx / 4)
            quarter = idx % 4
            ret.append(
                HourPrice(
                    day=self.dtmodel.hdate,
                    hour=hour,
                    quarter=quarter,
                    price=p,
                    passed=self._is_passed(hour, quarter),
                    hour_type=HourPrice.set_hour_type(
                        self.options.absolute_top_price, self.options.min_price, p
                    ),
                )
            )
        for idx, p in enumerate(prices_tomorrow):
            assert isinstance(p, (float, int))
            hour = int(idx / 4)
            quarter = idx % 4
            ret.append(
                HourPrice(
                    day=self.dtmodel.hdate_tomorrow,
                    hour=hour,
                    quarter=quarter,
                    price=p,
                    passed=False,
                    hour_type=HourPrice.set_hour_type(
                        self.options.absolute_top_price, self.options.min_price, p
                    ),
                )
            )
        await self.async_set_permittance(ret)
        return ret

    async def async_create_hour_prices_hourly(
        self, prices: list[float], prices_tomorrow: list[float] = []
    ) -> list:
        ret = []
        # todo: handle here first if prices or prices_tomorrow are 23 or 25 in length (dst shift)
        for idx, p in enumerate(prices):
            assert isinstance(p, (float, int))
            ret.append(
                HourPrice(
                    day=self.dtmodel.hdate,
                    hour=idx,
                    price=p,
                    passed=True if idx < self.dtmodel.hour else False,
                    hour_type=HourPrice.set_hour_type(
                        self.options.absolute_top_price, self.options.min_price, p
                    ),
                )
            )
        for idx, p in enumerate(prices_tomorrow):
            assert isinstance(p, (float, int))
            ret.append(
                HourPrice(
                    day=self.dtmodel.hdate_tomorrow,
                    hour=idx,
                    price=p,
                    passed=False,
                    hour_type=HourPrice.set_hour_type(
                        self.options.absolute_top_price, self.options.min_price, p
                    ),
                )
            )
        await self.async_set_permittance(ret)
        return ret

    async def async_set_permittance(self, hour_prices: list[HourPrice]) -> None:
        prices = normalize_prices([hp.price for hp in hour_prices])
        set_initial_permittance(
            hour_prices,
            self._set_price_mean(prices, self.model.adjusted_average),
            stdev(prices),
        )
        set_scooped_permittance(hour_prices, self.options.cautionhour_type_enum)
        self._offset_dict = self._set_offset_dict(prices, hour_prices[0].day)

    def _get_offset_dict(self) -> dict:
        vals = self._offset_dict.keys()
        if len(vals) == 1:
            return {TODAY: self._offset_dict.values(), TOMORROW: {}}
        elif len(vals) == 2:
            return {
                TODAY: self._offset_dict[min(vals)],
                TOMORROW: self._offset_dict[max(vals)],
            }
        # todo: log something here.
        return {
            TODAY: {},
            TOMORROW: {},
        }

    @staticmethod
    def _set_offset_dict(prices: list[float], day: date) -> dict:
        ret = {}
        _len = len(prices)
        today: list = []
        tomorrow: list = []
        match _len:
            case 23 | 24 | 25 | 92 | 96 | 100:
                today = prices
            case 47 | 48 | 49 | 188 | 192 | 196:
                today = prices[: len(prices) // 2]
                tomorrow = prices[len(prices) // 2 : :]
        ret[day] = deviation_from_mean(today)
        ret[day + timedelta(days=1)] = deviation_from_mean(tomorrow)
        # todo: handle interim
        return ret

    @staticmethod
    def _set_price_mean(prices: list[float], adjusted_average: float | None) -> float:
        # print(f"adj: {adjusted_average}")
        if not adjusted_average:
            return mean(prices)
        return mean([adjusted_average, mean(prices)])



from __future__ import annotations
from typing import TYPE_CHECKING
from datetime import datetime, timedelta

if TYPE_CHECKING:
    from .hourselection_service import HourSelectionService
from typing import Tuple
from .models.max_min_model import MaxMinModel

MINIMUM_DIFFERENCE = 0.1


class MaxMinCharge:
    def __init__(self, service: HourSelectionService, min_price: float | None) -> None:
        self.model = MaxMinModel(min_price=min_price)
        self.parent = service
        self.active: bool = False

    @property
    def total_charge(self) -> float:
        return self.model.total_charge

    @property
    def average_price(self) -> float | None:
        return self.model.average_price

    @property
    def original_total_charge(self) -> float:
        return sum(
            [
                hp.permittance * self.model.expected_hourly_charge
                for hp in self.parent.future_hours
            ]
        )

    @property
    def non_hours(self) -> list:
        print(self.parent.dtmodel.dt)
        return [
            k
            for k, v in self.model.input_hours.items()
            if v[1] == 0 and k >= self.parent.dtmodel.dt
        ]

    @property
    def dynamic_caution_hours(self) -> dict:
        return {
            k: v[1]
            for k, v in self.model.input_hours.items()
            if 0 < v[1] < 1 and k >= self.parent.dtmodel.dt
        }

    async def async_allow_decrease(self, car_connected: bool | None = None) -> bool:
        if car_connected is not None:
            return all(
                [
                    not car_connected,
                    len([k for k, v in self.model.input_hours.items() if v[1] > 0])
                    != 1,
                ]
            )
        return len([k for k, v in self.model.input_hours.items() if v[1] > 0]) != 1

    async def async_update(
        self,
        avg24,
        peak,
        max_desired: float,
        session_energy: float | None = None,
        car_connected: bool | None = None,
    ) -> None:
        allow_decrease: bool = False
        if await self.async_allow_decrease(car_connected):
            allow_decrease = True
            await self.async_setup(max_charge=peak)
        _session = session_energy or 0
        _desired = max_desired - _session
        _avg24 = round((avg24 / 1000), 1)
        self.model.expected_hourly_charge = peak - _avg24
        await self.async_increase_decrease(_desired, _avg24, peak, allow_decrease)

    async def async_increase_decrease(
        self, desired, avg24, peak, allow_decrease: bool
    ) -> None:
        for i in range(len(self.model.original_input_hours.items())):
            _load = self.total_charge - desired
            if _load > MINIMUM_DIFFERENCE and allow_decrease:
                await self.async_decrease()
            elif _load < MINIMUM_DIFFERENCE * -1:
                expected_charge = (desired - self.total_charge) / (peak - avg24)
                await self.async_increase(expected_charge)
            if abs(_load) < MINIMUM_DIFFERENCE:
                break

    async def async_initial_charge(self, avg24, peak) -> float:
        _avg24 = round((avg24 / 1000), 1)
        self.model.expected_hourly_charge = peak - _avg24
        total = 24 * (peak - _avg24)  # todo: fix 24 to be dynamic
        total -= len(self.non_hours) * (peak - _avg24)
        total -= sum(self.dynamic_caution_hours.values()) * (peak - _avg24)
        return total

    async def async_sum_charge(self, avg24, peak) -> float:
        total = 0
        for k, v in self.model.input_hours.items():
            total += (peak - avg24) * v[1]
        return total

    async def async_decrease(self):
        max_key = max(
            self.model.input_hours,
            key=lambda k: self.model.input_hours[k][0]
            if self.model.input_hours[k][1] != 0
            and self.model.input_hours[k][0] > self.model.min_price
            else -1,
        )
        self.model.input_hours[max_key] = (self.model.input_hours[max_key][0], 0)

    async def async_increase(self, expected_charge):
        min_key = min(
            self.model.input_hours,
            key=lambda k: self.model.input_hours[k][0]
            if self.model.input_hours[k][1] < 1
            and self.model.original_input_hours[k][1] > 0
            else 999,
        )
        self.model.input_hours[min_key] = (
            self.model.input_hours[min_key][0],
            min(min(1, expected_charge), self.model.original_input_hours[min_key][1]),
        )

    async def async_setup(
        self,
        max_charge: float,
        non_hours: list | None = None,
        dynamic_caution_hours: dict | None = None,
        prices: list | None = None,
        prices_tomorrow: list | None = None,
    ) -> None:
        if max_charge == 0:
            self.active = False
            return
        hour = self.parent.dtmodel.hour
        dt = self.parent.dtmodel.dt
        _non_hours = self.parent.non_hours if non_hours is None else non_hours
        _dynamic_caution_hours = (
            self.parent.dynamic_caution_hours
            if dynamic_caution_hours is None
            else dynamic_caution_hours
        )
        _prices = self.parent.model.prices_today if prices is None else prices
        _prices_tomorrow = (
            self.parent.model.prices_tomorrow
            if prices_tomorrow is None
            else prices_tomorrow
        )
        ret_today, ret_tomorrow = await self.async_loop_nonhours(
            dt, _non_hours, _prices, _prices_tomorrow
        )
        await self.async_loop_caution_hours(
            dt,
            _dynamic_caution_hours,
            _prices,
            _prices_tomorrow,
            ret_today,
            ret_tomorrow,
        )
        await self.async_add_available_hours(
            dt, _prices, _prices_tomorrow, ret_today, ret_tomorrow
        )
        self.model.input_hours = self._sort_dicts(ret_today, ret_tomorrow)
        self.model.original_input_hours = self.model.input_hours.copy()
        self.active = True

    @staticmethod
    async def async_add_available_hours(
        dt: datetime,
        prices: list,
        prices_tomorrow: list,
        ret_today: dict,
        ret_tomorrow: dict,
    ) -> None:
        _hour = dt.hour
        _range = (
            len(prices) - dt.hour + len(prices_tomorrow)
            if len(prices_tomorrow) > 0
            else len(prices) - dt.hour
        )

        for i in range(_range):
            if _hour < dt.hour and _hour not in ret_tomorrow.keys():
                ret_tomorrow[
                    (dt + timedelta(days=1))
                    .replace(hour=_hour)
                    .replace(minute=0)
                    .replace(second=0)
                ] = (
                    prices_tomorrow[_hour],
                    1,
                )
                # todo: must test this without valid prices tomorrow.
            elif _hour >= dt.hour and _hour not in ret_today.keys():
                ret_today[dt.replace(hour=_hour)] = (prices[_hour], 1)
            _hour += 1
            if _hour > 23:
                _hour = 0

    @staticmethod
    async def async_loop_caution_hours(
        dt: datetime,
        caution_hours: dict,
        prices: list,
        prices_tomorrow: list,
        ret_today: dict,
        ret_tomorrow: dict,
    ) -> None:
        for k, v in caution_hours.items():
            if k.day == dt.day and k.hour >= dt.hour:
                ret_today[k] = (prices[k.hour], v)
            elif len(prices_tomorrow) > 0:
                ret_tomorrow[k] = (prices_tomorrow[k.hour], v)

    @staticmethod
    async def async_loop_nonhours(
        dt: datetime, non_hours: list, prices: list, prices_tomorrow: list
    ) -> Tuple[dict, dict]:
        ret_today = {}
        ret_tomorrow = {}
        for n in non_hours:
            if n.day == dt.day and n.hour >= dt.hour:
                ret_today[n] = (prices[n.hour], 0)
            elif len(prices_tomorrow) > 0:
                ret_tomorrow[n] = (prices_tomorrow[n.hour], 0)
        return ret_today, ret_tomorrow

    @staticmethod
    def _sort_dicts(ret_today: dict, ret_tomorrow: dict) -> dict:
        ret = {}
        for k in sorted(ret_today.keys()):
            ret[k] = ret_today[k]
        for k in sorted(ret_tomorrow.keys()):
            ret[k] = ret_tomorrow[k]
        return ret



import logging
from statistics import mean
from typing import Tuple
from ...models.hourselection.cautionhourtype import CautionHourType
from ...models.hourselection.hourselection_model import HourSelectionModel
from ...models.hourselection.hourselection_options import HourSelectionOptions
from ..hoursselection_service_new.hourselection_service import HourSelectionService
from .hourselectionservice.hoursselection_helpers import convert_none_list

_LOGGER = logging.getLogger(__name__)


class Hoursselection:
    def __init__(
        self,
        absolute_top_price: float = 0,
        min_price: float = 0,
        cautionhour_type: str | CautionHourType = CautionHourType.SUAVE.value,
        blocknocturnal: bool = False,
        base_mock_hour: int | None = None,
    ):
        self.cautionhour_type_enum = (
            CautionHourType(cautionhour_type.lower())
            if isinstance(cautionhour_type, str)
            else cautionhour_type
        )
        self.model = HourSelectionModel(
            options=HourSelectionOptions(
                cautionhour_type=CautionHourType.get_num_value(cautionhour_type),
                cautionhour_type_enum=self.cautionhour_type_enum,
                min_price=min_price,
                top_price=absolute_top_price,
                blocknocturnal=blocknocturnal,
            )
        )
        self.model.validate()
        self.service = HourSelectionService(options=self.model.options)

    @property
    def offsets(self) -> dict:
        ret = {}
        ret["today"] = self.service.offset_dict["today"]
        ret["tomorrow"] = self.service.offset_dict["tomorrow"]
        return ret

    @property
    def non_hours(self) -> list:
        if self.service.max_min.active:
            ret = self.service.max_min.non_hours
        else:
            ret = self.internal_non_hours
        return sorted(ret)

    @property
    def caution_hours(self) -> list:
        return [k.hour for k, v in self.service.dynamic_caution_hours.items()]

    @property
    def dynamic_caution_hours(self) -> dict:
        if self.service.max_min.active:
            ret = self.service.max_min.dynamic_caution_hours
        else:
            ret = self.internal_dynamic_caution_hours
        keys = list(ret.keys())
        keys.sort()
        return {k: ret[k] for k in keys}

    @property
    def internal_non_hours(self) -> list:
        self.service.update()
        return self.service.non_hours

    @property
    def internal_dynamic_caution_hours(self) -> dict:
        self.service.update()
        return self.service.dynamic_caution_hours

    @property
    def prices(self) -> list:
        return self.model.prices_today

    @prices.setter
    def prices(self, val):
        self.model.prices_today = val
        if self.prices == self.prices_tomorrow:
            self.prices_tomorrow = []

    @property
    def prices_tomorrow(self) -> list:
        return self.model.prices_tomorrow

    @prices_tomorrow.setter
    def prices_tomorrow(self, val):
        ret = convert_none_list(val)
        self.model.prices_tomorrow = ret

    @property
    def adjusted_average(self):
        return self.model.adjusted_average

    @adjusted_average.setter
    def adjusted_average(self, val):
        if val != self.model.adjusted_average:
            self.model.adjusted_average = val

    async def async_update_adjusted_average(self, val):
        # self.adjusted_average = val
        await self.service.async_update_adjusted_average(val)
        # await self.async_update_prices(self.prices, self.prices_tomorrow)

    async def async_update_top_price(self, dyn_top_price=None) -> None:
        await self.model.options.async_set_absolute_top_price(dyn_top_price)
        await self.async_update_prices(self.prices, self.prices_tomorrow)

    async def async_update_prices(self, prices: list, prices_tomorrow: list = []):
        max_min: bool = False
        if any([self.prices != prices, self.prices_tomorrow != prices_tomorrow]):
            max_min = True
        self.prices = prices
        self.prices_tomorrow = prices_tomorrow
        await self.service.async_update_prices(self.prices, self.prices_tomorrow)

    async def async_get_average_kwh_price(self) -> Tuple[float | None, float | None]:
        ret_dynamic = None
        if self.service.max_min.active:
            ret_dynamic = self.service.max_min.average_price
            ret_static = self.service.average_kwh_price
            return ret_static, ret_dynamic
        else:
            ret_static = mean(
                [hp.permittance * hp.price for hp in self.service.future_hours]
            )
            try:
                return round(ret_static, 2), ret_dynamic
            except ZeroDivisionError as e:
                _LOGGER.warning(
                    f"get_average_kwh_price_core could not be calculated: {e}"
                )
            return 0, None

    async def async_get_total_charge(
        self, currentpeak: float
    ) -> Tuple[float, float | None]:
        ret_dynamic = None
        if self.service.max_min.active:
            ret_dynamic = self.service.max_min.total_charge
        self.model.current_peak = currentpeak
        ret_static = sum(
            [hp.permittance * currentpeak for hp in self.service.future_hours]
        )
        return round(ret_static, 1), ret_dynamic
